import arrow
import csv
import datetime
from enum import Enum
import ics.icalendar
import json
import logging
import pathlib
import os
import sys
import typer
from typing_extensions import Annotated

from nytid.signup import hr
from nytid.signup import sheets
import operator

from nytid.cli import courses as coursescli
from nytid.cli.signupsheets import SIGNUPSHEET_URL_PATH
from nytid import courses as courseutils
from nytid import schedules as schedutils
from nytid.signup import hr
from nytid.signup import sheets

import appdirs

try:
    import canvasapi
    import canvaslms.cli

    dirs_canvas = appdirs.AppDirs("canvaslms", "dbosk@kth.se")

    canvaslms_config = canvaslms.cli.read_configuration(
        f"{dirs_canvas.user_config_dir}/config.json"
    )

    CANVAS_SERVER, CANVAS_TOKEN = canvaslms.cli.login.load_credentials(canvaslms_config)

    if CANVAS_SERVER and CANVAS_TOKEN:
        canvas_session = canvasapi.Canvas(
            os.environ["CANVAS_SERVER"], os.environ["CANVAS_TOKEN"]
        )
    else:
        canvas_session = None
        logging.warning("Can't load Canvas credentials, run `canvaslms login`")
except ImportError as err:
    logging.warning(f"Can't import Canvas: {err}")
    canvas_session = None
except Exception as err:
    logging.warning(f"Can't load Canvas credentials: {err}")
    canvas_session = None
try:
    import ladok3
    import ladok3.cli

    dirs_ladok = appdirs.AppDirs("ladok", "dbosk@kth.se")

    LADOK_INST, LADOK_VARS = ladok3.cli.load_credentials(
        "f{dirs_ladok.user_config_dir}/config.json"
    )

    if LADOK_INST and LADOK_VARS:
        ladok_session = ladok3.LadokSession(LADOK_INST, LADOK_VARS)
    else:
        ladok_session = None
        logging.warning("Can't load LADOK credentials, run `ladok login`")
except ImportError as err:
    logging.warning(f"Can't import ladok3, not using LADOK data: {err}")
    ladok_session = None
except Exception as err:
    logging.warning(f"Can't load LADOK credentials: {err}")
    ladok_session = None
import cachetools
from nytid.signup import sheets
from nytid.signup import hr
import re
import typerconf

AMANUENSIS_CONTRACT_PATH = "amanuensis.contract_path"

cli = typer.Typer(name="hr", help="Manage sign-up sheets for teaching")


@cachetools.cached(cache={})
def get_canvas_courses(course_regex):
    """
    Returns a list of Canvas course objects matching the given course_regex.
    """
    courses = list(canvaslms.cli.courses.filter_courses(canvas_session, course_regex))
    return courses


@cachetools.cached(cache={})
def get_canvas_users(username_regex, course_regex):
    """
    Returns a list of Canvas user objects matching the given username_regex.
    Searches for username_regex in the courses matching course_regex.
    """
    courses = get_canvas_courses(course_regex)
    users = list(canvaslms.cli.users.filter_users(courses, username_regex))
    return users


def get_canvas_user(username, course_regex):
    """
    Takes a username and returns a Canvas user object.
    Searches for username in the courses matching course_regex.
    """
    users = get_canvas_users(".*", course_regex)
    username = username.strip()
    for user in users:
        if user.login_id.split("@")[0] == username or user.login_id == username:
            return user
    raise ValueError(f"Can't find {username} in Canvas")


def get_ladok_user(canvas_user):
    """
    Takes a Canvas user object and returns a LADOK student object.
    """
    try:
        return ladok_session.get_student(canvas_user.integration_id)
    except KeyError as err:
        raise KeyError(f"can't look up {canvas_user} in LADOK: {err}")


def to_hours(td):
    return td.total_seconds() / 60 / 60


def push_forward(start, end, push_start):
    """
    Takes a start and end date and pushes them forward so that start becomes
    push_start.
    """
    if push_start > start:
        end += push_start - start
        start = push_start

    return start, end


try:
    default_username = os.environ["USER"]
except KeyError:
    default_username = None

username_opt = typer.Option(
    help="Username to filter sign-up sheet for, "
    "defaults to logged in user's username."
)
detailed_opt = typer.Option(help="Output detailed user data.")
course_summary_opt = typer.Option(help="Print a summary of the course.")
amanuensis_summary_opt = typer.Option(help="Print a summary of the " "amanuensis.")
hourly_summary_opt = typer.Option(help="Print a summary of the hourly TAs.")
user_regex_opt = typer.Option(
    "--user", help="Regex to match TAs' usernames that " "should be included."
)
start_date_opt = typer.Option(
    help="The start date (inclusive, <=), "
    "when unset includes "
    "everything in the sign-up sheet. "
    "Set this to decide what to include from "
    "the sign-up sheet.",
    formats=["%Y-%m-%d"],
)
end_date_opt = typer.Option(
    help="The end date (not inclusive, <), "
    "when unset includes "
    "everything in the sign-up sheet. "
    "Set this to decide what to include from "
    "the sign-up sheet.",
    formats=["%Y-%m-%d"],
)
event_summary_opt = typer.Option(help="Print a summary of the hours per event " "type.")
push_start_opt = typer.Option(
    help="Push the dates of the contract so that it "
    "starts at this date. "
    "This keeps the same percentage.",
    formats=["%Y-%m-%d"],
)
set_start_opt = typer.Option(
    help="Force the start date of the contract to "
    "this date. Might modify percentage.",
    formats=["%Y-%m-%d"],
)
set_end_opt = typer.Option(
    help="Force the end date of the contract to " "this date. Might modify percentage.",
    formats=["%Y-%m-%d"],
)


@cli.command()
def users(
    course_regex: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    detailed: Annotated[bool, detailed_opt] = False,
):
    """
    Prints the list of all usernames booked on the course.
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    for user in hr.hours_per_TA(booked):
        user_obj = user

        if detailed:
            try:
                user_obj = get_canvas_user(user, course_regex)
            except Exception as err:
                logging.warning(f"Can't look up {user} in Canvas: {err}")
            else:
                try:
                    user_obj = get_ladok_user(user_obj)
                except Exception as err:
                    logging.warning(
                        f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                    )
                    pass
        print(user_obj)


@cli.command()
def time(
    course_regex: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    detailed: Annotated[bool, detailed_opt] = False,
    course_summary: Annotated[bool, course_summary_opt] = True,
    amanuensis_summary: Annotated[bool, amanuensis_summary_opt] = True,
    hourly_summary: Annotated[bool, hourly_summary_opt] = True,
):
    """
    Summarizes the time spent on teaching the course(s).
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    csvout = csv.writer(sys.stdout, delimiter="\t")
    if course_summary:
        h_per_student = hr.hours_per_student(booked)

        for event, hours in h_per_student.items():
            csvout.writerow([event, to_hours(hours), "h/student"])

        csvout.writerow(
            [
                "Booked (h)",
                to_hours(hr.total_hours(booked)),
                to_hours(hr.max_hours(booked)),
            ]
        )
    if amanuensis_summary:
        if course_summary:
            csvout.writerow([])
        if hourly_summary:
            csvout.writerow(["# Amanuensis"])

        amanuensis = hr.compute_amanuensis_data(booked)

        for user, data in amanuensis.items():
            if not user:
                continue
            user_obj = user

            if detailed:
                try:
                    user_obj = get_canvas_user(user, course_regex)
                except Exception as err:
                    logging.warning(f"Can't look up {user} in Canvas: {err}")
                else:
                    try:
                        user_obj = get_ladok_user(user_obj)
                    except Exception as err:
                        logging.warning(
                            f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                        )
                        pass
            csvout.writerow(
                [
                    user_obj,
                    f"{data[2]:.2f} h",
                    f"{100*hr.compute_percentage(*data):.1f}%",
                    f"{data[0].format('YYYY-MM-DD')}",
                    f"{data[1].format('YYYY-MM-DD')}",
                ]
            )
    if hourly_summary:
        if amanuensis_summary:
            csvout.writerow([])
            csvout.writerow(["# Hourly"])
        elif course_summary:
            csvout.writerow([])

        for user, hours in hr.hours_per_TA(booked).items():
            if not user or user in amanuensis:
                continue
            user_obj = user

            if detailed:
                try:
                    user_obj = get_canvas_user(user, course_regex)
                except Exception as err:
                    logging.warning(f"Can't look up {user} in Canvas: {err}")
                else:
                    try:
                        user_obj = get_ladok_user(user_obj)
                    except Exception as err:
                        logging.warning(
                            f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                        )
                        pass
            csvout.writerow([user_obj, to_hours(hours), "h"])


amanuensis = typer.Typer(name="amanuensis", help="Manage amanuensis employment")
cli.add_typer(amanuensis)


@amanuensis.command(name="create")
def amanuens_cmd(
    user_regex: Annotated[str, user_regex_opt] = ".*",
    start: Annotated[datetime.datetime, start_date_opt] = None,
    end: Annotated[datetime.datetime, end_date_opt] = None,
    push_start: Annotated[datetime.datetime, push_start_opt] = None,
    set_start: Annotated[datetime.datetime, set_start_opt] = None,
    set_end: Annotated[datetime.datetime, set_end_opt] = None,
    course_regex: Annotated[str, coursescli.course_arg_regex] = ".*",
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    detailed: Annotated[bool, detailed_opt] = True,
    event_summary: Annotated[bool, event_summary_opt] = False,
):
    """
    Computes amanuensis data for a TA.
    """
    if start:
        start = start.astimezone()
    if push_start:
        push_start = push_start.astimezone()
    if set_start:
        set_start = set_start.astimezone()
    if set_end:
        set_end = set_end.astimezone()
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course_regex, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    booked = sheets.filter_events_by_date(booked, start, end)

    amanuensis = hr.compute_amanuensis_data(booked, begin_date=start, end_date=end)

    user_pattern = re.compile(user_regex)
    first_print = True
    csvout = csv.writer(sys.stdout, delimiter="\t")
    path = pathlib.Path("./")
    try:
        path = pathlib.Path(typerconf.get(AMANUENSIS_CONTRACT_PATH))
    except KeyError as err:
        logging.warning(
            f"Can't find {AMANUENSIS_CONTRACT_PATH} in config, "
            f"storing contract data in `{path}`. Set by running "
            f"`nytid config {AMANUENSIS_CONTRACT_PATH} -s <path>`."
        )
    for user in amanuensis:
        if not user_pattern.match(user):
            continue
        if first_print:
            first_print = False
        else:
            print("\n")

        data = amanuensis[user]

        start = data[0]
        end = data[1]
        hours = data[2]
        if push_start:
            push_start = arrow.Arrow(push_start.year, push_start.month, push_start.day)
            start, end = push_forward(start, end, push_start)
        data = list(data)
        if set_start:
            start = data[0] = set_start
        if set_end:
            end = data[1] = set_end

        user_obj = user

        if detailed:
            try:
                user_obj = get_canvas_user(user, course_regex)
            except Exception as err:
                logging.warning(f"Can't look up {user} in Canvas: {err}")
            else:
                try:
                    user_obj = get_ladok_user(user_obj)
                except Exception as err:
                    logging.warning(
                        f"Can't look up {user} ({user_obj}) in LADOK: {err}"
                    )
                    pass

        row = [
            user_obj,
            start.date(),
            end.date(),
            f"{round(100*hr.compute_percentage(*data))}%",
        ]

        if event_summary:
            row.append(f"{hours:.2f} h")

        csvout.writerow(row)

        events = sheets.filter_events_by_TA(user, booked)
        events = filter(lambda x: user in sheets.get_booked_TAs_from_csv(x)[0], booked)
        events = list(
            map(lambda x: x[0 : len(sheets.SIGNUP_SHEET_HEADER)] + [user], events)
        )
        if event_summary:
            for event, hours in hr.hours_per_event(events).items():
                csvout.writerow([event, to_hours(hours), "h"])

        filename = f"{user}.{datetime.datetime.now().isoformat()}.json"

        path.mkdir(parents=True, exist_ok=True)

        with open(path / filename, "w") as outfile:
            json.dump(
                {
                    "user": user,
                    "start": start.isoformat() if start else None,
                    "set_start": set_start.isoformat() if set_start else None,
                    "push_start": push_start.isoformat() if push_start else None,
                    "end": end.isoformat() if end else None,
                    "set_end": set_end.isoformat() if set_end else None,
                    "events": events,
                },
                outfile,
                indent=2,
            )



import arrow
import canvasapi
import canvaslms.cli
import csv
import config
from nytid.signup import hr
from nytid.signup.hr import timesheet
from nytid.signup import sheets
import ladok3.kth
import operator
import os
import sys

def get_ladok_id(user, students):
    """
    Translates user ID to LADOK ID.
    - `user` is the username.
    - `students` is a list of students from Canvas (canvasapi.student.Student)
    """
    for student in students:
        if student.login_id == user:
            return student.integration_id
    raise KeyError(f"can't find {user}")

def to_hours(td):
    """
    Converts timedelta to hours.
    """
    return td.total_seconds()/60/60

def summarize_user(user, course, course_events,
                   salary=150):
    """
    Returns events where TA worked.
    - `user` is the TA's username.
    - `course` is the course code.
    - `course_events` is a list of events.
    - Optional `salary` is the hourly salary.
    """
    hours = to_hours(hr.hours_per_TA(course_events)[user])

    start_idx = sheets.SIGNUP_SHEET_HEADER.index("Start")
    end_idx = sheets.SIGNUP_SHEET_HEADER.index("End")
    type_idx = sheets.SIGNUP_SHEET_HEADER.index("Event")

    events = sheets.filter_events_by_TA(user, sorted(course_events,
            key=operator.itemgetter(start_idx)))
    events = filter(lambda x: user in sheets.get_booked_TAs_from_csv(x)[0], 
                    events)
    events = list(map(lambda x: x[0:len(sheets.SIGNUP_SHEET_HEADER)] + [user], 
                      events))

    xl_events = []

    for event in events:
        end = arrow.get(event[end_idx])
        start = arrow.get(event[start_idx])
        event_type = event[type_idx]
        time = end-start
        time_with_prep = to_hours(hr.round_time(
                                hr.add_prep_time(time, event_type,
                                                 date=start.date())))

        xl_events.append({
                "datum": str(start.date()),
                "tid": str(start.time()),
                "kurskod": course,
                "typ": event_type,
                "timmar": to_hours(time),
                "koeff": hr.prep_factor(event_type, date=start.date(),
                                        amanuensis=False),
                "omr_tid": time_with_prep,
                "belopp": time_with_prep * salary
            })

    return xl_events


def main():
    """
    Main command code
    """
    cs = canvasapi.Canvas(os.environ["CANVAS_SERVER"], 
                          os.environ["CANVAS_TOKEN"])
    ls = ladok3.kth.LadokSession(os.environ["KTH_LOGIN"],
                                 os.environ["KTH_PASSWD"])

    course = next(canvaslms.cli.courses.filter_courses(cs, "datintro22"))
    students = list(course.get_users())

    course_events = []

    for course, url in config.SIGNUP.items():
        course_events += sheets.read_signup_sheet_from_url(url)

    if len(sys.argv) > 1:
        users = sys.argv[1:]
    else:
        users = hr.hours_per_TA(course_events)

    for user in users:
        try:
            ta = ls.get_student(get_ladok_id(f"{user}@kth.se", students))
        except KeyError as err:
            print(f"can't look up {user} in LADOK: {err}")
            continue

        ta_events = summarize_user(user, course, course_events)
        try:
            timesheet.make_xlsx(ta.personnummer,
                                f"{ta.first_name} {ta.last_name}",
                                f"{user}@kth.se",
                                ta_events,
                                course_leader=("Daniel Bosk", "dbosk@kth.se"),
                                HoD="Karl Meinke",
                                org="JH", project="1102",
                                course_leader_signature="~/Pictures/signature.png")
        except AttributeError as err:
            print(f"can't access {user}'s LADOK data: {err}")
            continue

if __name__ == "__main__":
    main()



import datetime
from enum import Enum
import ics.icalendar
import logging
import pathlib
import sys
import typer
from typing_extensions import Annotated

from nytid.cli import courses as coursescli
from nytid import courses as courseutils
from nytid import schedules as schedutils
from nytid.signup import hr
from nytid.signup import sheets

from nytid.signup import sheets
from nytid.signup import utils
import os, platform, subprocess

SIGNUPSHEET_URL_PATH = "signupsheet.url"

cli = typer.Typer(name="signupsheets", help="Manage sign-up sheets for teaching")

outpath_opt = typer.Option(
    help="Path where to write the sign-up sheet "
    "files. Default is in each course's "
    "data path."
)
edit_opt = typer.Option(help="If specified, opens each generated sheet " "for editing.")
url_arg = typer.Argument(
    help="The URL for the sign-up sheet. "
    "For Google Sheets, it's the same URL as the "
    "one shared with TAs to sign up. "
    "For others, it's a URL to a CSV file."
)


@cli.command(name="generate")
def generate_signup_sheet(
    course: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    outpath: Annotated[pathlib.Path, outpath_opt] = None,
    edit: Annotated[bool, edit_opt] = False,
):
    """
    Generates a sign-up sheet to be used with Google Sheets or similar for TAs to
    sign up.
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)

    now = datetime.datetime.now()
    timestamp = now.strftime("%Y%m%d-%H%M")
    for (course, register), course_conf in courses.items():
        num_students = course_conf.get("num_students")
        num_groups = course_conf.get("num_groups")

        group_size = round(num_students / num_groups)

        def needed_TAs(event):
            return utils.needed_TAs(event, group_size=group_size)

        if not outpath:
            try:
                data_root = courseutils.get_course_data(course, register)
            except KeyError as err:
                logging.warning(err)
            except PermissionError as err:
                logging.warning(
                    f"You don't have access to {course} in {register}: {err}"
                )
                outfile = f"./signup-{course}-{timestamp}.csv"
                logging.warning(f"Writing file to current working directory: {outfile}")
            else:
                outfile = data_root.path / f"signup-{course}-{timestamp}.csv"
        else:
            outfile = outpath / f"signup-{course}-{timestamp}.csv"

        url = course_conf.get("ics")
        sheets.generate_signup_sheet(outfile, url, needed_TAs)
        if edit:
            the_os = platform.system()
            if the_os == "Darwin":
                subprocess.call(["open", outfile])
            elif the_os == "Windows":
                os.startfile(outfile)
            else:
                subprocess.call(["xdg-open", outfile])


@cli.command()
def set_url(
    course: Annotated[str, coursescli.course_arg_regex],
    url: Annotated[str, url_arg],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
):
    """
    Sets the URL of the sign-up sheet for the course(s).
    """
    registers = coursescli.registers_regex(register)
    courses = {}
    for course_reg in coursescli.courses_regex(course, registers):
        try:
            courses[course_reg] = courseutils.get_course_config(*course_reg)
        except KeyError as err:
            logging.warning(err)
        except PermissionError as err:
            course, register = course_reg
            logging.warning(f"You don't have access to {course} in {register}: {err}")
    if not courses:
        sys.exit(1)
    for _, conf in courses.items():
        conf.set(SIGNUPSHEET_URL_PATH, url)



"""The nytid course management module"""

import typerconf

import pathlib
from nytid.courses import registry
from nytid import storage
import typing


def new(
    name: str,
    register: str = None,
    kwdata: dict = None,
    config: typerconf.Config = typerconf,
):
    """
    Creates a new course. It takes the following arguments:

    - `name` (mandatory), which is the human readable nickname of the course. This
      is used to refer to the course with other parts of nytid.

    - `register` is the name of the course register to use. Required if there are
      more than one course register in the config. Default is `None`, which selects
      the only available course register. If more registers, raises exception
      `ValueError`.

    - `kwdata` is a dictionary containing key-value pairs for the course
      configuration. `kwdata` expect the following:

        - `contact` contains the contact information for the course responsible, it can
          be any format, but we recommend "Firstname Lastname <email>". The default
          value is built from values in the main config file:

            - `me.name` contains the name,
            - `me.email` contains the email address.

        - `code`, which is the course code. This is to relate the course to other
          courses through `related_codes`. However, it's not mandatory.

        - `related_codes`, a list of related course codes. Courses with one of these
          course codes can share TAs.

        - `ics` (optional, default None), a URL to an ICS file with the schedule of the
          course. E.g. a URL to a TimeEdit export/subscription.

        - `data_path` is the path to the course's data directory. If not supplied,
          we append `/data` to the course's config directory.


      Any other key--value pairs will simply be written to the course configuration
      as is.

    The default `config` is the default config of the `typerconf` package.
    """
    if not register:
        registers = registry.ls(config=config)

        if len(registers) > 1:
            raise ValueError(f"Must specify a course register: {registers}")
        elif len(registers) < 1:
            raise ValueError(f"There are no course registers in the config.")
        else:
            register = registers[0]

    register_path = registry.get(register, config=config)

    with storage.open_root(f"{register_path}/{name}") as root:
        try:
            with root.open("config.json", "r") as course_conf_file:
                pass
        except FileNotFoundError:
            course_conf = typerconf.Config()
        else:
            raise FileExistsError(
                f"The config for {name} already exists: "
                f"{register_path}/{name}/config.json"
            )

        if not kwdata:
            kwdata = {}

        try:
            contact = kwdata["contact"]
        except KeyError:
            contact = None

        if not contact:
            try:
                contact = config.get("me.name")
            except KeyError:
                contact = ""

            try:
                email = config.get("me.email")
            except KeyError:
                pass
            else:
                if contact:
                    contact += f" <{email}>"
                else:
                    contact = email

        kwdata["contact"] = contact
        if "code" not in kwdata:
            kwdata["code"] = None
        if "related_codes" not in kwdata:
            kwdata["related_codes"] = None
        if "ics" not in kwdata:
            kwdata["ics"] = None
        try:
            data_path = kwdata["data_path"]
        except KeyError:
            data_path = None

        if not data_path:
            data_path = str(root.path / "data")

        kwdata["data_path"] = data_path

        for key, value in kwdata.items():
            course_conf.set(key, value)

        with root.open("config.json", "w") as course_conf_file:
            course_conf.write_config(course_conf_file)


def get_course_config(
    course: str, register: str = None, config: typerconf.Config = typerconf
) -> typerconf.Config:
    """
    Returns a typerconf.Config object for the course's configuration. Writeback
    is enabled meaning that any changes made to the config object will be written
    back to the config file.

    `course` identifies the course in the `register`. If `register` is None,
    search through all registers in the registry, use the first one found
    (undefined in which order duplicates are sorted).

    The default `config` is the default config of the `typerconf` package.
    """
    if register:
        register_map = {register: registry.get(register, config=config)}
    else:
        register_map = {
            register: registry.get(register, config=config)
            for register in registry.ls(config=config)
        }
    try:
        conf_path = get_course_conf_path(course, register_map)
    except KeyError as err:
        raise KeyError(
            f"Couldn't uniquely identify {course} in "
            f"registers {register_map.keys()}: {err}"
        )

    conf = typerconf.Config(conf_file=conf_path)
    return conf


def get_course_conf_path(course, register_map):
    """
    Find the course named `course` among all the courses in the registers found
    in `register_map`, a list of (name, path)-tuples.

    If `course` is not a unique match, it raises a `KeyError`.
    """
    hits_from_register_dirs = []
    conf_path = None

    for register_name, register_path in register_map.items():
        courses = all_courses(register_path)
        matching_courses = list(filter(lambda x: x == course, courses))
        if len(matching_courses) > 0:
            conf_path = f"{register_path}/{course}/config.json"
            hits_from_register_dirs.append(register_name)

    if not conf_path:
        raise KeyError(f"Couldn't find course {course}.")
    elif len(hits_from_register_dirs) > 1:
        raise KeyError(
            f"Course `{course}` found in "
            f"several course registers: {hits_from_register_dirs}."
        )

    return conf_path


def all_courses(register_path: pathlib.Path) -> typing.List[str]:
    """
    Returns a list (generator) of all courses found in `register_path`.
    """
    for file in pathlib.Path(register_path).iterdir():
        if file.is_dir():
            yield file.name


def get_course_data(
    course: str, register: str = None, config: typerconf.Config = typerconf
) -> storage.StorageRoot:
    """
    Returns a StorageRoot object for the data directory of the course `course`.

    If `register` is `None` (default), the first occurence of `course` in all
    registers will be used.
    """
    course_conf = get_course_config(course, register, config)
    return storage.open_root(course_conf.get("data_path"))
